from .editor import PresentationSectionType, get_config, set_config
from .config import Config

set_config(Config)

# NORMAL = PresentationSectionType.NORMAL
# SKIP = PresentationSectionType.SKIP
# LOOP = PresentationSectionType.LOOP
# COMPLETE_LOOP = PresentationSectionType.COMPLETE_LOOP



import jinja2
import pathlib
import os
from distutils.dir_util import copy_tree
from typing import List

from .config import get_config
from .scene import Section

BASE_DIR = pathlib.Path(__file__).parent.absolute()
ENV = jinja2.Environment(
    loader=jinja2.FileSystemLoader([
        os.path.join(BASE_DIR, "..", "app", "templates"),
        os.path.join(BASE_DIR, "..", "app", "main", "templates"),
        os.path.join(BASE_DIR, "..", "app", "error", "templates"),
    ]),
    autoescape=jinja2.select_autoescape(["html"])
)


def emulate_url_for(endpoint: str, path: str = "", name: str = "", filename: str = "") -> str:
    """Translate flask urls into local urls for project.
    ``name`` is being ignored as the output path will be from the project dir already.
    """
    if endpoint == "static":
        return filename
    elif endpoint == "main.serve_project_static":
        return path
    raise ValueError(f"Endpoint '{endpoint}' can't be emulated.")


def export_presentation(project_name: str, sections: List[Section]) -> None:
    html = ENV.get_template("edit_project.html").render(
        present_export=True,
        version=get_config().VERSION,
        name=project_name,
        sections=sections,
        url_for=emulate_url_for
    )
    with open(os.path.join(project_name, "index.html"), "w") as file:
        file.write(html)
    copy_tree(os.path.join(BASE_DIR, "..", "app", "static", "webpack"), os.path.join(project_name, "webpack"))
    copy_tree(os.path.join(BASE_DIR, "..", "app", "static", "img"), os.path.join(project_name, "img"))



"""Load a Manim Editor project."""
import os
from fractions import Fraction
from typing import Any, Dict, List, Optional, Tuple

from .manim_loader import Section, valid_json_load
from .commands import walk
from .config import get_config


def get_project_section(raw_section: Dict[str, Any]) -> Section:
    """Create :class:`.Section` from dict read from a project json file created by the Manim Editor."""

    return Section(
        int(raw_section["id"]),
        raw_section["name"],
        # TODO: should be converted to PresentationSectionType
        raw_section["type"],
        raw_section["original_video"],
        int(raw_section["width"]),
        int(raw_section["height"]),
        Fraction(raw_section["fps"]),
        float(raw_section["duration"]),
        raw_section["project_name"],
        raw_section["in_project_video"],
        raw_section["in_project_thumbnail"],
        int(raw_section["in_project_id"]),
    )


def get_project(path: str) -> Tuple[Optional[str], List[Section]]:
    """Parse project json file if valid.
    ``path`` is path to prject dir, not to the json index file.
    Otherwise return ``None`` as name.
    """
    raw_sections = valid_json_load(os.path.join(path, "project.json"), get_config().PROJECT_SCHEMA)
    if raw_sections is None:
        return None, []
    name = os.path.basename(path)
    sections: List[Section] = []
    for raw_section in raw_sections:
        sections.append(get_project_section(raw_section))
    return name, sections


def get_projects() -> Dict[str, List[Section]]:
    """Get all projects in the current working directory.
    No recursive search will be applied."""
    # get all projects in this dir
    project_paths: List[str] = []
    for root, _, files in walk(".", 1):
        for file in files:
            if file == "project.json":
                project_paths.append(root)
    projects: Dict[str, List[Section]] = {}
    for project_path in project_paths:
        name, sections = get_project(project_path)
        # don't include invalid
        if name is not None:
            projects[name] = sections
    return projects



import os
import click
import logging
from waitress import serve
import socket
from typing import Optional

from .app import create_app
from .config import Config
from .editor import set_config, create_project_dir, get_scene, populate_project_with_loaded_sections, export_presentation


def find_open_port():
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(("", 0))
    sock.listen(1)
    _, port = sock.getsockname()
    sock.close()

    return port


def run_normal() -> None:
    set_config(Config)
    app = create_app(Config)
    # disable logging for every get/post request
    log = logging.getLogger("werkzeug")
    log.setLevel(logging.ERROR)
    port = find_open_port()
    url = f"http://localhost:{port}"
    print(f"Starting Manim Editor v{Config.VERSION}. Open {url} in a browser. (Press CTRL+C to quit)")
    serve(app, host="0.0.0.0", port=port)


def run_debug() -> None:
    Config.set("FFMPEG_LOGLEVEL", "info")
    set_config(Config)
    app = create_app(Config)
    app.run(debug=True)


# TODO: quick test, this is only for testing purposes
def run_quick_present_export(section_index_path: str) -> None:
    """Create a project form a single scene and export the presentation."""
    Config.set("FFMPEG_LOGLEVEL", "info")
    set_config(Config)
    project_name = os.path.basename(section_index_path)[:-5]

    success, msg = create_project_dir(project_name)
    if not success:
        raise RuntimeError(msg)
    scene = get_scene(section_index_path, 0)
    if scene is None:
        raise RuntimeError(f"Couldn't find a section index file at '{section_index_path}'.")
    if not populate_project_with_loaded_sections(project_name, scene.sections):
        raise RuntimeError("Failed to populate project.")
    export_presentation(project_name, scene.sections)


@click.command()
@click.option("--debug", is_flag=True, help="Launch in Debug mode.")
@click.option("--version", is_flag=True, help="Print version and exit.")
@click.option("--quick_present_export", help="Create a project from a scene and export the presentation. The path to the section index file must be provided as the argument.")
def main(debug: bool, version: bool, quick_present_export: Optional[str]) -> None:
    if version:
        print(f"Manim Editor v{Config.VERSION}")
        return
    if quick_present_export is not None:
        run_quick_present_export(quick_present_export)
        return
    if debug:
        run_debug()
    else:
        run_normal()


if __name__ == "__main__":
    main()



"""Functions used to load the render results of Manim."""
import os
from pathlib import Path
from typing import List, Optional, Any, Dict
from fractions import Fraction

from .scene import Section, Scene
from .config import get_config
from .commands import walk, valid_json_load


def get_manim_section(raw_section: Dict[str, Any], index_path: str, new_id: int) -> Section:
    """Create :class:`.Section` from dict read from a section index created by Manim."""

    parent_path = Path(index_path).parent.absolute()
    return Section(
        new_id,
        raw_section["name"],
        raw_section["type"],
        os.path.join(parent_path, raw_section["video"]),
        int(raw_section["width"]),
        int(raw_section["height"]),
        Fraction(raw_section["avg_frame_rate"]),
        float(raw_section["duration"]),
    )


def get_scene(path: str, new_id: int) -> Optional[Scene]:
    """Check if path points to a valid section index file.
    Return ``None`` if file isn't a section index.
    """
    raw_sections = valid_json_load(path, get_config().SECTION_INDEX_SCHEMA)
    if raw_sections is None:
        return None

    sections: List[Section] = []
    for raw_section in raw_sections:
        sections.append(get_manim_section(raw_section, path, len(sections)))
    return Scene(
        new_id,
        os.path.basename(path)[:-5],
        os.path.abspath(path),
        os.path.getmtime(path),
        sections)


def get_scenes() -> List[Scene]:
    """Search recursively in ``path`` for any valid JSON section index files."""
    scene_index_paths: List[str] = []
    for root, _, files in walk(".", get_config().RECURSION_DEPTH):
        for file in files:
            if file.endswith(".json"):
                scene_index_paths.append(os.path.join(root, file))

    scenes: List[Scene] = []
    for scene_index_path in scene_index_paths:
        scene = get_scene(scene_index_path, len(scenes))
        if scene is not None:
            scenes.append(scene)
    return scenes



import os
import shutil
import time
import pathlib
from fractions import Fraction
from enum import Enum
from typing import List, Dict, Any

from .commands import run_ffmpeg


class PresentationSectionType(str, Enum):
    # start, end, wait for continuation by user
    NORMAL = "presentation.normal"
    # start, end, immediately continue to next section
    SKIP = "presentation.skip"
    # start, end, restart, immediately continue to next section when continued by user
    LOOP = "presentation.loop"
    # start, end, restart, finish animation first when user continues
    COMPLETE_LOOP = "presentation.complete_loop"


class Section:
    """Representation of Manim :class:`.Section`.

    Attributes
    ----------
    id
        Unique id for this section.
    name
        Human readable, non-unique name for this section.
    type
        How should this section be played?
    original_video
        Path to original video file.
    width
        width of the video
    height
        Height of the video.
    fps
        Frame rate of the video as :class:`.Fraction`.
    duration
        Duration of the video in seconds.
    project_name
        Name of the project this section is used in.
    in_project_video
        Path to copied video file relative to project file.
    in_project_thumbnail
        Path to thumbnail file relative to project file.
    in_project_id
        Id for this section that is unique in its project.

    See Also
    --------
    :class:`.PresentationSectionType`
    """

    def __init__(self,
                 id: int,
                 name: str,
                 type: PresentationSectionType,
                 original_video: str,
                 width: int,
                 height: int,
                 fps: Fraction,
                 duration: float,
                 # only to be used when loading from project file
                 project_name: str = "",
                 in_project_video: str = "",
                 in_project_thumbnail: str = "",
                 in_project_id: int = -1):
        self.id = id
        self.name = name
        self.type = type
        self.original_video = original_video
        self.width = width
        self.height = height
        self.fps = fps
        self.duration = duration
        # to be set once project is being populated
        self.project_name = project_name
        self.in_project_video = in_project_video
        self.in_project_thumbnail = in_project_thumbnail
        self.in_project_id = in_project_id

    def get_in_project_video_abs(self) -> str:
        return os.path.join(self.project_name, self.in_project_video)

    def get_in_project_thumbnail_abs(self) -> str:
        return os.path.join(self.project_name, self.in_project_thumbnail)

    def set_project(self, project_name: str, in_project_id: int) -> bool:
        """Hand this video over to a project.
        Update attributes and copy assets.
        """
        self.project_name = project_name
        self.in_project_id = in_project_id
        # TODO support other file types as well
        self.in_project_video = f"video_{in_project_id:04}.mp4"
        self.in_project_thumbnail = f"thumb_{in_project_id:04}.jpg"

        return self.convert_video() and self.extract_thumbnail()

    def convert_video(self) -> bool:
        """Convert original video into fragmented format that can be played by the video player.
        Also copy it to project dir.
        Return False at failure.
        """
        print(f"Converting video to '{self.in_project_video}' for section '{self.name}'.")
        if not run_ffmpeg([
            "-i",
            self.original_video,
            "-movflags",
            "frag_keyframe+empty_moov+default_base_moof",
            self.get_in_project_video_abs(),
            "-y",
        ]):
            print("Video Conversion failed.")
            return False
        return True

    def extract_thumbnail(self) -> bool:
        """Create thumbnail for section in project dir.
        Return False at failure.
        """
        print(f"Extracting thumbnail '{self.in_project_thumbnail}' from video for section '{self.name}'.")
        if not run_ffmpeg([
            "-sseof",
            "-3",
            "-i",
            self.original_video,
            "-update",
            "1",
            "-q:v",
            "1",
            self.get_in_project_thumbnail_abs(),
            "-y",
        ]):
            print("Thumbnail extraction failed.")
            return False
        return True

    def get_dict(self) -> Dict[str, Any]:
        """Get dictionary representation."""
        return {
            "id": self.id,
            "name": self.name,
            "type": self.type,
            "original_video": self.original_video,
            "width": self.width,
            "height": self.height,
            "fps": str(self.fps),
            "duration": self.duration,
            "project_name": self.project_name,
            "in_project_video": self.in_project_video,
            "in_project_thumbnail": self.in_project_thumbnail,
            "in_project_id": self.in_project_id,
        }


class Scene:
    """Representation of the entire section index of one scene.

    Attributes
    ----------
    id
        unique id for this scene
    name
        name for the represented scene
    path
        path to index JSON file
    last_modified
        seconds since the epoch of last modification of index file
    sections
        list of sections in scene
    """

    def __init__(self, id: int, name: str, path: str, last_modified: float, sections: List[Section]):
        self.id = id
        self.name = name
        self.path = path
        self.last_modified = last_modified
        self.sections = sections

    def get_last_modified(self) -> str:
        return time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(self.last_modified))

    def get_rel_dir_path(self) -> str:
        parent_path = pathlib.Path(self.path).parent.absolute()
        return os.path.relpath(parent_path)
